/** Classes needed for the lexer
*/
module antlrrt.lexing;

// private import std.stream;

private import antlrrt.utils;
private import antlrrt.exceptions;
private import antlrrt.collections;


// import java.io.InputStream;


/** A token is minimally a token type.  Subclasses can add the text matched
 *  for the token and line info.
 */
public class Token {   // implements Cloneable
    // constants
    public static const int MIN_USER_TYPE = 4;
    public static const int NULL_TREE_LOOKAHEAD = 3;
    public static const int INVALID_TYPE = 0;
    public static const int EOF_TYPE = 1;
    public static const int EOF = 1;
    public static const int SKIP = -1;

    // each Token has at least a token type
    protected int type = INVALID_TYPE;

    // the illegal token object
    public static Token badToken;
    static this() {badToken = new Token(INVALID_TYPE, asWString("<no text>"));}

    public this() {
    }

    public this(int t) {
        type = t;
    }

    public this(int t, WString txt) {
        type = t;
        setText(txt);
    }

    public int getColumn() {
        return 0;
    }

    public int getLine() {
        return 0;
    }

    public WString getFilename() {
        return null;
    }

    public void setFilename(WString name) {
    }

    public WString getText() {
        return asWString("<no text>");
    }

    public void setText(WString t) {
    }

    public void setColumn(int c) {
    }

    public void setLine(int l) {
    }

    public int getType() {
        return type;
    }

    public void setType(int t) {
        type = t;
    }

    public WString toWString() {
        return "[\"" ~ getText() ~ "\",<" ~ asWString(getType()) ~ ">]";
    }
}



public class CommonToken : Token {
    // most tokens will want line and text information
    protected int line;
    protected WString text = null;
    protected int col;

    public this() {
    }

    public this(int t, WString txt) {
        type = t;
        setText(txt);
    }

    public this(WString s) {
        text = s;
    }
    
    /// Needed for D
    public static Token antlrFactory()
    {
        return new CommonToken();
    }

    public int getLine() {
        return line;
    }

    public WString getText() {
        return text;
    }

    public void setLine(int l) {
        line = l;
    }

    public void setText(WString s) {
        text = s;
    }

    public WString toWString() {
        return "[\"" ~ getText() ~ "\",<" ~ asWString(type) ~ ">,line=" ~ asWString(line) ~ ",col=" ~ asWString(col) ~ "]";
    }

    /** Return token's start column */
    public int getColumn() {
        return col;
    }

    public void setColumn(int c) {
        col = c;
    }
}



public class CommonHiddenStreamToken : CommonToken {
    protected CommonHiddenStreamToken hiddenBefore;
    protected CommonHiddenStreamToken hiddenAfter;

    public this() {
        super();
    }

    public this(int t, WString txt) {
        super(t, txt);
    }

    public this(WString s) {
        super(s);
    }

    public CommonHiddenStreamToken getHiddenAfter() {
        return hiddenAfter;
    }

    public CommonHiddenStreamToken getHiddenBefore() {
        return hiddenBefore;
    }

    protected void setHiddenAfter(CommonHiddenStreamToken t) {
        hiddenAfter = t;
    }

    protected void setHiddenBefore(CommonHiddenStreamToken t) {
        hiddenBefore = t;
    }
}


/** A circular buffer object used by CharBuffer */
alias Queue!(wchar) CharQueue;


/**A Stream of characters fed to the lexer from a InputStream that can
 * be rewound via mark()/rewind() methods.
 * <p>
 * A dynamic array is used to buffer up all the input characters.  Normally,
 * "k" characters are stored in the buffer.  More characters may be stored during
 * guess mode (testing syntactic predicate), or when LT(i>k) is referenced.
 * Consumption of characters is deferred.  In other words, reading the next
 * character is not done by conume(), but deferred until needed by LA or LT.
 * <p>
 *
 * @see CharQueue
 */
public abstract class InputBuffer {
    // Number of active markers
    protected int nMarkers = 0;

    // Additional offset used when markers are active
    protected int markerOffset = 0;

    // Number of calls to consume() since last LA() or LT() call
    protected int numToConsume = 0;

    // Circular queue
    protected CharQueue queue;

    /** Create an input buffer */
    public this() {
        queue = new CharQueue(1);
    }

    /** This method updates the state of the input buffer so that
     *  the text matched since the most recent mark() is no longer
     *  held by the buffer.  So, you either do a mark/rewind for
     *  failed predicate or mark/commit to keep on parsing without
     *  rewinding the input.
     */
    public void commit() {
        nMarkers--;
    }

    /** Mark another character for deferred consumption */
    public void consume() {
        numToConsume++;
    }

    /** Ensure that the input buffer is sufficiently full */
    public abstract void fill(int amount); // throws CharStreamException;

    public WString getLAChars() {
        StringBuffer la = new StringBuffer();
        for (int i = markerOffset; i < queue.nbrEntries; i++)
            la.append(queue.elementAt(i));
        return la.toWString();
    }

    public WString getMarkedChars() {
        StringBuffer marked = new StringBuffer();
        for (int i = 0; i < markerOffset; i++)
            marked.append(queue.elementAt(i));
        return marked.toWString();
    }

    public boolean isMarked() {
        return (nMarkers != 0);
    }

    /** Get a lookahead character */
    public wchar LA(int i)  {   // throws CharStreamException
        fill(i);
        return queue.elementAt(markerOffset + i - 1);
    }

    /**Return an integer marker that can be used to rewind the buffer to
     * its current state.
     */
    public int mark() {
        syncConsume();
        nMarkers++;
        return markerOffset;
    }

    /**Rewind the character buffer to a marker.
     * @param mark Marker returned previously from mark()
     */
    public void rewind(int mark) {
        syncConsume();
        markerOffset = mark;
        nMarkers--;
    }

    /** Reset the input buffer
     */
    public void reset() {
        nMarkers = 0;
        markerOffset = 0;
        numToConsume = 0;
        queue.reset();
    }

    /** Sync up deferred consumption */
    protected void syncConsume() {
        while (numToConsume > 0) {
            if (nMarkers > 0) {
                // guess mode -- leave leading characters and bump offset.
                markerOffset++;
            }
            else {
                // normal mode -- remove first character
                queue.removeFirst();
            }
            numToConsume--;
        }
    }
}



/**A Stream of characters fed to the lexer from a InputStream that can
 * be rewound via mark()/rewind() methods.
 * <p>
 * A dynamic array is used to buffer up all the input characters.  Normally,
 * "k" characters are stored in the buffer.  More characters may be stored during
 * guess mode (testing syntactic predicate), or when LT(i>k) is referenced.
 * Consumption of characters is deferred.  In other words, reading the next
 * character is not done by conume(), but deferred until needed by LA or LT.
 * <p>
 *
 * @see antlr.CharQueue
 */

// SAS: Move most functionality into InputBuffer -- just the file-specific
//      stuff is in here



public class CharBuffer : InputBuffer {
    // wchar source
    /*transient*/ Reader input;

    /** Create a character buffer */
    public this(Reader input_) { // SAS: for proper text i/o
        super();
        input = input_;
    }

    /** Ensure that the character buffer is sufficiently full */
    public void fill(int amount) {    // throws CharStreamException
        try {
            syncConsume();
            // Fill the buffer sufficiently to hold needed characters
            while (queue.nbrEntries < amount + markerOffset) {
                // Append the next character
                queue.append(input.read());
            }
        }
        catch (IOException io) {
            throw new CharStreamIOException(io);
        }
    }
}



/**A Stream of characters fed to the lexer from a InputStream that can
 * be rewound via mark()/rewind() methods.
 * <p>
 * A dynamic array is used to buffer up all the input characters.  Normally,
 * "k" characters are stored in the buffer.  More characters may be stored during
 * guess mode (testing syntactic predicate), or when LT(i>k) is referenced.
 * Consumption of characters is deferred.  In other words, reading the next
 * character is not done by conume(), but deferred until needed by LA or LT.
 * <p>
 *
 * @see antlr.CharQueue
 */
// SAS: added this class to handle Binary input w/ FileInputStream

public class ByteBuffer : CharBuffer {

    // char source
    /*transient*/ InputStream input;


    /** Create a character buffer */
    public this(InputStream input_) {
        super(new BinaryReader(input_));
        // input = input_;
    }

    /* Ensure that the character buffer is sufficiently full
    public void fill(int amount) {    // throws CharStreamException
        try {
            syncConsume();
            // Fill the buffer sufficiently to hold needed characters
            while (queue.nbrEntries < amount + markerOffset) {
                // Append the next character
                queue.append(cast(wchar)input.getc());
            }
        }
        catch (IOException io) {
            throw new CharStreamIOException(io);
        }
    } */ 
}



/** This object contains the data associated with an
 *  input stream of characters.  Multiple lexers
 *  share a single LexerSharedInputState to lex
 *  the same input stream.
 */
public class LexerSharedInputState {
    protected int column = 1;
    protected int line = 1;
    protected int tokenStartColumn = 1;
    protected int tokenStartLine = 1;
    protected InputBuffer input;

    /** What file (if known) caused the problem? */
    protected WString filename;

    public int guessing = 0;

    public this(InputBuffer inbuf) {
        input = inbuf;
    }

    public this(InputStream inp) {
        this(new ByteBuffer(inp));
    }

    public this(Reader inp) {
        this(new CharBuffer(inp));
    }

    public void reset() {
        column = 1;
        line = 1;
        tokenStartColumn = 1;
        tokenStartLine = 1;
        guessing = 0;
        filename = null;
        input.reset();
    }
}


public interface TokenStream {
    public Token nextToken();   //  throws TokenStreamException;
}


/** A token stream MUX (multiplexor) knows about n token streams
 *  and can multiplex them onto the same channel for use by token
 *  stream consumer like a parser.  This is a way to have multiple
 *  lexers break up the same input stream for a single parser.
 *    Or, you can have multiple instances of the same lexer handle
 *  multiple input streams; this works great for includes.
 */
public class TokenStreamSelector : /*implements*/ TokenStream {
    /** The set of inputs to the MUX */
    protected Hashtable!(WString, TokenStream) inputStreamNames;

    /** The currently-selected token stream input */
    protected TokenStream input;

    /** Used to track stack of input streams */
    protected Stack!(TokenStream) streamStack;  // new LList();

    public this() {
        // super();
        inputStreamNames = new Hashtable!(WString, TokenStream)();
        streamStack = new Vector!(TokenStream)();
    }

    public void addInputStream(TokenStream stream, WString key) {
        inputStreamNames.put(key, stream);
    }

    /** Return the stream from tokens are being pulled at
     *  the moment.
     */
    public TokenStream getCurrentStream() {
        return input;
    }

    public TokenStream getStream(WString sname) {
        TokenStream stream = inputStreamNames.get(sname);
        if (stream is null) {
            throw new IllegalArgumentException("TokenStream " ~ sname ~ " not found");
        }
        return stream;
    }

    public Token nextToken() {   //  throws TokenStreamException 
        // return input.nextToken();
        // keep looking for a token until you don't
        // get a retry exception.
        for (; ;) {
            try {
                return input.nextToken();
            }
            catch (TokenStreamRetryException r) {
                // just retry "forever"
            }
        }
    }

    public TokenStream pop() {
        TokenStream stream = streamStack.pop();
        select(stream);
        return stream;
    }

    public void push(TokenStream stream) {
        streamStack.push(input); // save current stream
        select(stream);
    }

    public void push(WString sname) {
        streamStack.push(input);
        select(sname);
    }

    /** Abort recognition of current Token and try again.
     *  A stream can push a new stream (for include files
     *  for example, and then retry(), which will cause
     *  the current stream to abort back to this.nextToken().
     *  this.nextToken() then asks for a token from the
     *  current stream, which is the new "substream."
     */
    public void retry() {   //  throws TokenStreamRetryException 
        throw new TokenStreamRetryException();
    }

    /** Set the stream without pushing old stream */
    public void select(TokenStream stream) {
        input = stream;
    }

    public void select(WString sname)  {   // throws IllegalArgumentException
        input = getStream(sname);
    }
}



/** This object is a TokenStream that passes through all
 *  tokens except for those that you tell it to discard.
 *  There is no buffering of the tokens.
 */
public class TokenStreamBasicFilter: /*implements*/ TokenStream {
    /** The set of token types to discard */
    protected BitSet discardMask;

    /** The input stream */
    protected TokenStream input;

    public this(TokenStream input) {
        this.input = input;
        discardMask = new BitSet();
    }

    public void discard(int ttype) {
        discardMask.add(ttype);
    }

    public void discard(BitSet mask) {
        discardMask = mask;
    }

    public Token nextToken()  {    // throws TokenStreamException
        Token tok = input.nextToken();
        while (tok !is null && discardMask.member(tok.getType())) {
            tok = input.nextToken();
        }
        return tok;
    }
}



/**This object filters a token stream coming from a lexer
 * or another TokenStream so that only certain token channels
 * get transmitted to the parser.
 *
 * Any of the channels can be filtered off as "hidden" channels whose
 * tokens can be accessed from the parser.
 */
public class TokenStreamHiddenTokenFilter : TokenStreamBasicFilter, /*implements*/ TokenStream {
    // protected BitSet discardMask;
    protected BitSet hideMask;

    protected CommonHiddenStreamToken nextMonitoredToken;

    /** track tail of hidden list emanating from previous
     *  monitored token
     */
    protected CommonHiddenStreamToken lastHiddenToken;

    protected CommonHiddenStreamToken firstHidden = null;

    public this(TokenStream input) {
        super(input);
        hideMask = new BitSet();
    }

    protected void consume() {      // throws TokenStreamException
        nextMonitoredToken = cast(CommonHiddenStreamToken)input.nextToken();
    }

    private void consumeFirst() {    // throws TokenStreamException
        consume(); // get first token of input stream

        // Handle situation where hidden or discarded tokens
        // appear first inp input stream
        CommonHiddenStreamToken p = null;
        // while hidden or discarded scarf tokens
        while (hideMask.member(LA(1).getType()) || discardMask.member(LA(1).getType())) {
            if (hideMask.member(LA(1).getType())) {
                if (p is null) {
                    p = LA(1);
                }
                else {
                    p.setHiddenAfter(LA(1));
                    LA(1).setHiddenBefore(p); // double-link
                    p = LA(1);
                }
                lastHiddenToken = p;
                if (firstHidden is null) {
                    firstHidden = p; // record hidden token if first
                }
            }
            consume();
        }
    }

    public BitSet getDiscardMask() {
        return discardMask;
    }

    /** Return a ptr to the hidden token appearing immediately after
     *  token t inp the input stream.
     */
    public CommonHiddenStreamToken getHiddenAfter(CommonHiddenStreamToken t) {
        return t.getHiddenAfter();
    }

    /** Return a ptr to the hidden token appearing immediately before
     *  token t inp the input stream.
     */
    public CommonHiddenStreamToken getHiddenBefore(CommonHiddenStreamToken t) {
        return t.getHiddenBefore();
    }

    public BitSet getHideMask() {
        return hideMask;
    }

    /** Return the first hidden token if one appears
     *  before any monitored token.
     */
    public CommonHiddenStreamToken getInitialHiddenToken() {
        return firstHidden;
    }

    public void hide(int m) {
        hideMask.add(m);
    }

    public void hide(BitSet mask) {
        hideMask = mask;
    }

    protected CommonHiddenStreamToken LA(int i) {
        return nextMonitoredToken;
    }

    /** Return the next monitored token.
     *  Test the token following the monitored token.
     *  If following is another monitored token, save it
     *  for the next invocation of nextToken (like a single
     *  lookahead token) and return it then.
     *  If following is unmonitored, nondiscarded (hidden)
     *  channel token, add it to the monitored token.
     *
     *  Note: EOF must be a monitored Token.
     */
    public Token nextToken() {    // throws TokenStreamException
        // handle an initial condition; don't want to get lookahead
        // token of this splitter until first call to nextToken
        if (LA(1) is null) {
            consumeFirst();
        }

        // we always consume hidden tokens after monitored, thus,
        // upon entry LA(1) is a monitored token.
        CommonHiddenStreamToken monitored = LA(1);
        // point to hidden tokens found during last invocation
        monitored.setHiddenBefore(lastHiddenToken);
        lastHiddenToken = null;

        // Look for hidden tokens, hook them into list emanating
        // from the monitored tokens.
        consume();
        CommonHiddenStreamToken p = monitored;
        // while hidden or discarded scarf tokens
        while (hideMask.member(LA(1).getType()) || discardMask.member(LA(1).getType())) {
            if (hideMask.member(LA(1).getType())) {
                // attach the hidden token to the monitored in a chain
                // link forwards
                p.setHiddenAfter(LA(1));
                // link backwards
                if (p !is monitored) { //hidden cannot point to monitored tokens
                    LA(1).setHiddenBefore(p);
                }
                p = lastHiddenToken = LA(1);
            }
            consume();
        }
        return monitored;
    }
}



/** class implements a String-like object whose sole purpose is to be
  entered into a lexer HashTable.  It uses a lexer object to get
  information about case sensitivity.
*/

public class ANTLRHashString {
    private WString s;
    private CharScanner lexer;
    private static const int prime = 151;


    public this(wchar[] buf, int length, CharScanner lexer) {
        this.lexer = lexer;
        setBuffer(buf, length);
    }

    // Hash strings constructed this way are unusable until setBuffer or setString are called.
    public this(CharScanner lexer) {
        this.lexer = lexer;
    }

    public this(WString s, CharScanner lexer) {
        this.lexer = lexer;
        setString(s);
    }

    private final wchar charAt(int index) {
        return s.charAt(index);
    }

    // Return true if o is an ANTLRHashString equal to this.
    public boolean equals(Object o) {
        ANTLRHashString s = cast(ANTLRHashString) o;
        
        if (s is null)
        {
            WString w = cast(WString) o;
            if (w is null)
                return false;
            
            s = new ANTLRHashString(w, lexer);
        }

        return opCmp(s) == 0;   // TODO More efficient?
    }

    /// D hash function
    public uint toHash() {
        return cast(uint) hashCode();
    }
    
    int opCmp(Object o) {
        ANTLRHashString s = cast(ANTLRHashString) o;
        
        if (s is null)
            return super.opCmp(o);
        
        int l = length();
        if (l != s.length()) {
            return l - s.length();
        }
        if (lexer.getCaseSensitiveLiterals()) {
            for (int i = 0; i < l; i++) {
                if (charAt(i) != s.charAt(i)) {
                    return charAt(i) - s.charAt(i);
                }
            }
        }
        else {
            for (int i = 0; i < l; i++) {
                if (lexer.toLower(charAt(i)) != lexer.toLower(s.charAt(i))) {
                    return lexer.toLower(charAt(i)) - lexer.toLower(s.charAt(i));
                }
            }
        }
        return 0;
    }

    
    /// From Java
    public int hashCode() {
        int hashval = 0;
        int l = length();

        if (lexer.getCaseSensitiveLiterals()) {
            for (int i = 0; i < l; i++) {
                hashval = hashval * prime + charAt(i);
            }
        }
        else {
            for (int i = 0; i < l; i++) {
                hashval = hashval * prime + lexer.toLower(charAt(i));
            }
        }
        return hashval;
    }

    private final int length() {
        return s.length;
    }

    public void setBuffer(wchar[] buf, int buflength) {
        this.s = new WString(buf[0 .. buflength]);    // TODO More efficient?
    }

    public void setString(WString s) {
        this.s = s;
    }
}



/** 
  Factory function for Token construction
*/
typedef Token function() TokenFactoryFunc;



public abstract class CharScanner : /*implements*/ TokenStream {
    static const wchar NO_CHAR = 0;
    public static const wchar EOF_CHAR = wchar.init;  //cast(wchar)-1;
    
    /** Needed in D to map a classname to its TokenFactoryFunc.
    */
    public static TokenFactoryFunc[WString] classMap;
    
    public static void registerClass(WString className, TokenFactoryFunc cls)
    {
        classMap[className] = cls;
    }
    
    static this()
    {
        registerClass(asWString("antlrrt.CommonToken"), &CommonToken.antlrFactory);
    }
    
    /** Replaces Class.forName()
    */
    public static TokenFactoryFunc classForName(WString className)
    {
        return classMap[className];
    }

    
    protected ANTLRStringBuffer text; // text of current token

    protected boolean saveConsumedInput = true; // does consume() save characters?
    protected TokenFactoryFunc tokenObjectClass; // what kind of tokens to create?
    protected boolean caseSensitive = true;
    protected boolean caseSensitiveLiterals = true;
    protected Hashtable!(ANTLRHashString, int) literals; // set by subclass

    /** Tab chars are handled by tab() according to this value; override
     *  method to do anything weird with tabs.
     */
    protected int tabsize = 8;

    protected Token _returnToken = null; // used to return tokens w/o using return val.

    // Hash string used so we don't new one every time to check literals table
    protected ANTLRHashString hashString;

    protected LexerSharedInputState inputState;

    /** Used during filter mode to indicate that path is desired.
     *  A subsequent scan error will report an error as usual if
     *  acceptPath=true;
     */
    protected boolean commitToPath = false;

    /** Used to keep track of indentdepth for traceIn/Out */
    protected int traceDepth = 0;

    public this() {
        text = new ANTLRStringBuffer();
        hashString = new ANTLRHashString(this);
        setTokenObjectClass(asWString("antlrrt.CommonToken"));
    }

    public this(InputBuffer cb) { // SAS: use generic buffer
        this();
        inputState = new LexerSharedInputState(cb);
    }

    public this(LexerSharedInputState sharedState) {
        this();
        inputState = sharedState;
    }

    public void append(wchar c) {
        if (saveConsumedInput) {
            text.append(c);
        }
    }

    public void append(WString s) {
        if (saveConsumedInput) {
            text.append(s);
        }
    }

    public void commit() {
        inputState.input.commit();
    }

    public void consume() {    // throws CharStreamException
        if (inputState.guessing == 0) {
            wchar c = LA(1);
            if (caseSensitive) {
                append(c);
            }
            else {
                // use input.LA(), not LA(), to get original case
                // CharScanner.LA() would toLower it.
                append(inputState.input.LA(1));
            }
            if (c == '\t') {
                tab();
            }
            else {
                inputState.column++;
            }
        }
        inputState.input.consume();
    }

    /** Consume chars until one matches the given wchar */
    public void consumeUntil(int c) {    // throws CharStreamException
        while (LA(1) != EOF_CHAR && LA(1) != c) {
            consume();
        }
    }

    /** Consume chars until one matches the given set */
    public void consumeUntil(BitSet set) {   // throws CharStreamException
        while (LA(1) != EOF_CHAR && !set.member(LA(1))) {
            consume();
        }
    }

    public boolean getCaseSensitive() {
        return caseSensitive;
    }

    public final boolean getCaseSensitiveLiterals() {
        return caseSensitiveLiterals;
    }

    public int getColumn() {
        return inputState.column;
    }

    public void setColumn(int c) {
        inputState.column = c;
    }

    public boolean getCommitToPath() {
        return commitToPath;
    }

    public WString getFilename() {
        return inputState.filename;
    }

    public InputBuffer getInputBuffer() {
        return inputState.input;
    }

    public LexerSharedInputState getInputState() {
        return inputState;
    }

    public void setInputState(LexerSharedInputState state) {
        inputState = state;
    }

    public int getLine() {
        return inputState.line;
    }

    /** return a copy of the current text buffer */
    public WString getText() {
        return text.toWString();
    }

    public Token getTokenObject() {
        return _returnToken;
    }

    public wchar LA(int i) {   // throws CharStreamException
        if (caseSensitive) {
            return inputState.input.LA(i);
        }
        else {
            return toLower(inputState.input.LA(i));
        }
    }

    protected Token makeToken(int t) {
        try {
            Token tok = tokenObjectClass();
            tok.setType(t);
            tok.setColumn(inputState.tokenStartColumn);
            tok.setLine(inputState.tokenStartLine);
            // tracking real start line now: tok.setLine(inputState.line);
            return tok;
        }
        /* catch (InstantiationException ie) {
            panic("can't instantiate token: " + tokenObjectClass);
        }
        catch (IllegalAccessException iae) {
            panic("Token class is not accessible" + tokenObjectClass);
        } */
        catch (Exception e) {    // TODO
            panic(asWString("Token creation error"));
        }
        
        return Token.badToken;
    }

    public int mark() {
        return inputState.input.mark();
    }

    public void match(wchar c) {   // throws MismatchedCharException, CharStreamException
        if (LA(1) != c) {
            throw new MismatchedCharException(LA(1), c, false, this);
        }
        consume();
    }

    public void match(BitSet b) {   // throws MismatchedCharException, CharStreamException
        if (!b.member(LA(1))) {
            throw new MismatchedCharException(LA(1), b, false, this);
        }
        else {
            consume();
        }
    }

    public void match(WString s) {   // throws MismatchedCharException, CharStreamException
        int len = s.length();
        for (int i = 0; i < len; i++) {
            if (LA(1) != s.charAt(i)) {
                throw new MismatchedCharException(LA(1), s.charAt(i), false, this);
            }
            consume();
        }
    }
    
    public void match(wchar[] s) {
        int len = s.length;
        for (int i = 0; i < len; i++) {
            if (LA(1) != s[i]) {
                throw new MismatchedCharException(LA(1), s[i], false, this);
            }
            consume();
        }
    }
    

    public void matchNot(wchar c) {   // throws MismatchedCharException, CharStreamException
        if (LA(1) == c) {
            throw new MismatchedCharException(LA(1), c, true, this);
        }
        consume();
    }

    public void matchRange(wchar c1, wchar c2) {   // throws MismatchedCharException, CharStreamException
        if (LA(1) < c1 || LA(1) > c2) throw new MismatchedCharException(LA(1), c1, c2, false, this);
        consume();
    }

    public void newline() {
        inputState.line++;
        inputState.column = 1;
    }

    /** advance the current column number by an appropriate amount
     *  according to tab size. This method is called from consume().
     */
    public void tab() {
        int c = getColumn();
        int nc = ( ((c-1)/tabsize) + 1) * tabsize + 1;  // calculate tab stop
        setColumn( nc );
    }

    public void setTabSize( int size ) {
          tabsize = size;
    }

    public int getTabSize() {
        return tabsize;
    }

    /** @see #panic(WString)
     */
    public void panic() {
        System.err.println("CharScanner: panic");
        System.exit(1);
    }

    /** This method is executed by ANTLR internally when it detected an illegal
     *  state that cannot be recovered from.
     *  The default implementation of this method calls
     *  {@link java.lang.System.exit(int)} and writes directly to
     *  {@link java.lang.System.err)} , which is usually not appropriate when
     *  a translator is embedded into a larger application. <em>It is highly
     *  recommended that this method be overridden to handle the error in a
     *  way appropriate for your application (e.g. throw an unchecked
     *  exception)</em>.
     */
    public void panic(WString s) {
        System.err.println("CharScanner; panic: " ~ s);
        System.exit(1);
    }

    /** Parser error-reporting function can be overridden in subclass */
    public void reportError(RecognitionException ex) {
        System.err.println(ex.getMessage());
    }

    /** Parser error-reporting function can be overridden in subclass */
    public void reportError(WString s) {
        if (getFilename() is null) {
            System.err.println("error: " ~ s);
        }
        else {
            System.err.println(getFilename() ~ ": error: " ~ s);
        }
    }

    /** Parser warning-reporting function can be overridden in subclass */
    public void reportWarning(WString s) {
        if (getFilename() is null) {
            System.err.println("warning: " ~ s);
        }
        else {
            System.err.println(getFilename() ~ ": warning: " ~ s);
        }
    }

    public void resetText() {
        text.setLength(0);
        inputState.tokenStartColumn = inputState.column;
        inputState.tokenStartLine = inputState.line;
    }

    public void rewind(int pos) {
         inputState.input.rewind(pos);
         // RK: should not be here, it is messing up column calculation
         // setColumn(inputState.tokenStartColumn);
    }

    public void setCaseSensitive(boolean t) {
        caseSensitive = t;
    }

    public void setCommitToPath(boolean commit) {
        commitToPath = commit;
    }

    public void setFilename(WString f) {
        inputState.filename = f;
    }

    public void setLine(int line) {
        inputState.line = line;
    }

    public void setText(WString s) {
        resetText();
        text.append(s);
    }

    public void setTokenObjectClass(WString cl) {
        try {
            tokenObjectClass = classForName(cl);
        }
        catch (Exception ce) {    // TODO
            panic("Token class not found: " ~ cl);
        }
    }

    /** Test the token text against the literals table
      Override this method to perform a different literals test
    */
    public int testLiteralsTable(int ttype) {
        hashString.setBuffer(text.getBuffer(), text.length());
        int* literalsIndex = literals.getptr(hashString);
        if (literalsIndex !is null) {
            ttype = *literalsIndex;
        }
        return ttype;
    }

    /** Test the text passed in against the literals table
     * Override this method to perform a different literals test
     * This is used primarily when you want to test a portion of
     * a token.
     */
    public int testLiteralsTable(WString text, int ttype) {
        ANTLRHashString s = new ANTLRHashString(text, this);
        int* literalsIndex = literals.getptr(hashString);
        if (literalsIndex !is null) {
            ttype = *literalsIndex;
        }
        return ttype;
    }

    // Override this method to get more specific case handling
    public wchar toLower(wchar c) {
        return Character.toLowerCase(c);
    }


    public void traceIndent() {
        for (int i = 0; i < traceDepth; i++)
            System.outp.print(" ");
    }

    public void traceIn(WString rname) {  // throws CharStreamException
        traceDepth += 1;
        traceIndent();
        System.outp.println("> lexer " ~ rname ~ "; c==" ~ asWString(LA(1)));
    }


    public void traceOut(WString rname) {  // throws CharStreamException
        traceIndent();
        System.outp.println("< lexer " ~ rname ~ "; c==" ~ asWString(LA(1)));
        traceDepth -= 1;
    }

    /** This method is called by YourLexer.nextToken() when the lexer has
     *  hit EOF condition.  EOF is NOT a character.
     *  This method is not called if EOF is reached during
     *  syntactic predicate evaluation or during evaluation
     *  of normal lexical rules, which presumably would be
     *  an IOException.  This traps the "normal" EOF condition.
     *
     *  uponEOF() is called after the complete evaluation of
     *  the previous token and only if your parser asks
     *  for another token beyond that last non-EOF token.
     *
     *  You might want to throw token or char stream exceptions
     *  like: "Heh, premature eof" or a retry stream exception
     *  ("I found the end of this file, go back to referencing file").
     */
    public void uponEOF() {   // throws TokenStreamException, CharStreamException
    }
}


